/* $Id:   Exp $ */
/******************************************************************************
*
*	XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
*	AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
*	SOLUTIONS FOR XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE,
*	OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
*	APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
*	THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
*	AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
*	FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
*	WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
*	IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
*	REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
*	INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*	FOR A PARTICULAR PURPOSE.
*
*	 Copyright (C) 2006 Vreelin Engineering, Inc.  All Rights Reserved.
*	(c) Copyright  2007 Xilinx Inc.
*	All rights reserved.
*
******************************************************************************/
/*****************************************************************************/
/**
*
* @file xusb_l.h
*
* This header file contains identifiers and low-level driver function prototypes
* (or macros) that can be used to access the USB device. High-level driver
* function prototypes are defined in xusb_l.h.
*
* <pre>
*
* MODIFICATION HISTORY:
*
* Ver   Who  Date     Changes
* ----- ---- -------- ---------------------------------------------------------
* 1.00a hvm   2/22/07 First release
*
* </pre>
*
******************************************************************************/
#ifndef XUSB_L_H		/* prevent circular inclusions */
#define XUSB_L_H		/* by using protection macros */

#ifdef __cplusplus
extern "C" {
#endif

/***************************** Include Files *********************************/

#include "xbasic_types.h"
#include "xio.h"

/************************** Constant Definitions *****************************/

#include "xparameters.h"
#define XPAR_XUSB_NUM_INSTANCES 1
#define XPAR_USB_0_DEVICE_ID	0
#define XPAR_USB_0_BASEADDR	XPAR_OPB_USB2_DEVICE_0_BASEADDR

/** @name Register Map
 *
 * Register offsets for the USB device.
 * @{
 */
#define XUSB_EP0_CONFIG_OFFSET		0x0000 /**< EP0 Config Reg Offset */
#define XUSB_SETUP_PKT_ADDR_OFFSET	0x0080 /**< Setup Packet Address */
#define XUSB_ADDRESS_OFFSET		0x4000 /**< Address Register */
#define XUSB_CONTROL_OFFSET		0x4004 /**< Control Register */
#define XUSB_STATUS_OFFSET		0x4008 /**< Status Register */
#define XUSB_FRAMENUM_OFFSET 		0x400C	/**< Frame Number Register */
#define XUSB_IER_OFFSET 		0x4010	/**< Interrupt Enable Reg */
#define XUSB_BUFFREADY_OFFSET		0x4014	/**< Buffer Ready Register */
#define XUSB_TESTMODE_OFFSET		0x4018	/**< Test Mode Register */

/* @} */

/** @name Address Register
 *
 * @{
 */
#define XUSB_ADDRESS_ADDR_MASK		0x7F	/**< Address Mask  */

/* @} */

/** @name Control Register
 *
 * @{
 */
#define XUSB_CONTROL_USB_READY_MASK 	0x80000000 /**< USB ready Mask */

/* @} */

/** @name Status Register and Interrupt Enable Register
 *
 * <b> Status Register  </b>
 *
 * This register holds the status flags for the USB device.
 *
 * <b> Interrupt Enable Register  </b>
 *
 * This register is used to enable interrupt sources for the USB device.
 * Writing a '1' to a bit in this register enables the corresponding Interrupt.
 * Writing a '0' to a bit in this register disables the corresponding Interrupt.
 *
 * Both these registers have the same bit definitions and are only defined
 * once.
 * @{
 */
#define XUSB_STATUS_GLOBAL_INTR_MASK	0x80000000 /**< Global Intr Enable */
#define XUSB_STATUS_RESET_MASK		0x00800000 /**< USB Reset Mask */
#define XUSB_STATUS_SUSPEND_MASK	0x00400000 /**< USB Suspend Mask */
#define XUSB_STATUS_DISCONNECT_MASK	0x00200000 /**< USB Disconnect Mask */
#define XUSB_STATUS_FIFO_BUFF_RDY_MASK	0x00100000 /**< FIFO Buff Ready Mask */
#define XUSB_STATUS_FIFO_BUFF_FREE_MASK	0x00080000 /**< FIFO Buff Free Mask */
#define XUSB_STATUS_SETUP_PACKET_MASK	0x00040000 /**< Setup packet received */
#define XUSB_STATUS_SOF_PACKET_MASK	0x00020000 /**< SOF packet received */
#define XUSB_STATUS_HIGH_SPEED_MASK	0x00010000 /**< USB Speed Mask */
#define XUSB_STATUS_EP7_BUFF2_COMP_MASK	0x00008000 /**< EP 7 Buff 2 Processed */
#define XUSB_STATUS_EP6_BUFF2_COMP_MASK	0x00004000 /**< EP 6 Buff 2 Processed */
#define XUSB_STATUS_EP5_BUFF2_COMP_MASK	0x00002000 /**< EP 5 Buff 2 Processed */
#define XUSB_STATUS_EP4_BUFF2_COMP_MASK	0x00001000 /**< EP 4 Buff 2 Processed */
#define XUSB_STATUS_EP3_BUFF2_COMP_MASK	0x00000800 /**< EP 3 Buff 2 Processed */
#define XUSB_STATUS_EP2_BUFF2_COMP_MASK	0x00000400 /**< EP 2 Buff 2 Processed */
#define XUSB_STATUS_EP1_BUFF2_COMP_MASK	0x00000200 /**< EP 1 Buff 2 Processed */
#define XUSB_STATUS_EP7_BUFF1_COMP_MASK	0x00000080 /**< EP 7 Buff 1 Processed */
#define XUSB_STATUS_EP6_BUFF1_COMP_MASK	0x00000040 /**< EP 6 Buff 1 Processed */
#define XUSB_STATUS_EP5_BUFF1_COMP_MASK	0x00000020 /**< EP 5 Buff 1 Processed */
#define XUSB_STATUS_EP4_BUFF1_COMP_MASK	0x00000010 /**< EP 4 Buff 1 Processed */
#define XUSB_STATUS_EP3_BUFF1_COMP_MASK	0x00000008 /**< EP 3 Buff 1 Processed */
#define XUSB_STATUS_EP2_BUFF1_COMP_MASK	0x00000004 /**< EP 2 Buff 1 Processed */
#define XUSB_STATUS_EP1_BUFF1_COMP_MASK	0x00000002 /**< EP 1 Buff 1 Processed */
#define XUSB_STATUS_EP0_BUFF1_COMP_MASK	0x00000001 /**< EP 0 Buff 1 Processed */

#define XUSB_STATUS_EP_BUFF2_SHIFT	8	   /**< EP buffer offset */
/* @} */

/** @name Frame Number Register
 *
 * @{
 */
#define XUSB_FRAMENUM_FRAME_MASK	0x00007FF8 /**< Frame Number Mask  */
#define XUSB_FRAMENUM_MICRO_FRAME_MASK	0x00000007 /**< Micro Frame Mask */
#define XUSB_FRAMENUM_FRAME_SHIFT	0x3	   /**< Frame Number Shift */

/* @} */

/** @name Buffer Ready Register
 *
 * @{
 */
#define XUSB_BUFFREADY_EP7_BUFF2_MASK	0x00008000  /**< EP 7 Buff 2 Ready */
#define XUSB_BUFFREADY_EP6_BUFF2_MASK	0x00004000  /**< EP 6 Buff 2 Ready */
#define XUSB_BUFFREADY_EP5_BUFF2_MASK	0x00002000  /**< EP 5 Buff 2 Ready */
#define XUSB_BUFFREADY_EP4_BUFF2_MASK	0x00001000  /**< EP 4 Buff 2 Ready */
#define XUSB_BUFFREADY_EP3_BUFF2_MASK	0x00000800  /**< EP 3 Buff 2 Ready */
#define XUSB_BUFFREADY_EP2_BUFF2_MASK	0x00000400  /**< EP 2 Buff 2 Ready */
#define XUSB_BUFFREADY_EP1_BUFF2_MASK	0x00000200  /**< EP 1 Buff 2 Ready */
#define XUSB_BUFFREADY_EP7_BUFF1_MASK	0x00000080  /**< EP 7 Buff 1 Ready */
#define XUSB_BUFFREADY_EP6_BUFF1_MASK	0x00000040  /**< EP 6 Buff 1 Ready */
#define XUSB_BUFFREADY_EP5_BUFF1_MASK	0x00000020  /**< EP 5 Buff 1 Ready */
#define XUSB_BUFFREADY_EP4_BUFF1_MASK	0x00000010  /**< EP 4 Buff 1 Ready */
#define XUSB_BUFFREADY_EP3_BUFF1_MASK	0x00000008  /**< EP 3 Buff 1 Ready */
#define XUSB_BUFFREADY_EP2_BUFF1_MASK	0x00000004  /**< EP 2 Buff 1 Ready */
#define XUSB_BUFFREADY_EP1_BUFF1_MASK	0x00000002  /**< EP 1 Buff 1 Ready */
#define XUSB_BUFFREADY_EP0_BUFF_MASK	0x00000001  /**< EP 0 Buffer Ready */

#define XUSB_STATUS_INTR_EVENT_MASK	(XUSB_STATUS_RESET_MASK | \
					XUSB_STATUS_SUSPEND_MASK | \
					XUSB_STATUS_DISCONNECT_MASK | \
					XUSB_STATUS_SOF_PACKET_MASK)
#define XUSB_STATUS_INTR_BUFF_COMP_ALL_MASK  \
					(XUSB_STATUS_EP7_BUFF2_COMP_MASK | \
					XUSB_STATUS_EP6_BUFF2_COMP_MASK |  \
					XUSB_STATUS_EP5_BUFF2_COMP_MASK |  \
					XUSB_STATUS_EP4_BUFF2_COMP_MASK |  \
					XUSB_STATUS_EP3_BUFF2_COMP_MASK |  \
					XUSB_STATUS_EP2_BUFF2_COMP_MASK |  \
					XUSB_STATUS_EP1_BUFF2_COMP_MASK |  \
					XUSB_STATUS_EP7_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP6_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP5_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP4_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP3_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP2_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP1_BUFF1_COMP_MASK |  \
					XUSB_STATUS_EP0_BUFF1_COMP_MASK )

#define XUSB_STATUS_INTR_ALL_MASK	0x80FFFEFF /**< All Interrupts Mask */

/* @} */

/** @name Test Mode Register
 *
 * @{
 */
#define XUSB_TESTMODE_J_MASK		0x00000001 /**< Test Mode J */
#define XUSB_TESTMODE_K_MASK            0x00000002 /**< Test Mode K */
#define XUSB_TESTMODE_SE0_NAK_MASK	0x00000003 /**< Test Mode SE0 NAK */
#define XUSB_TESTMODE_TEST_PKT_MASK	0x00000004 /**< Test Packet Mode */

/* @} */

/** @name Endpoint Configuration Space offsets
 *
 * @{
 */
#define XUSB_EP_CFGSTATUS_OFFSET	0x00	/**< Endpoint Config Status  */
#define XUSB_EP_BUF0COUNT_OFFSET	0x08	/**< Buffer 0 Count */
#define XUSB_EP_BUF1COUNT_OFFSET	0x0C	/**< Buffer 1 Count */

/* @} */

/** @name Endpoint Configuration Status Register
 *
 * @{
 */
#define XUSB_EP_CFG_VALID_MASK		0x80000000 /**< Endpoint Valid bit */
#define XUSB_EP_CFG_STALL_MASK		0x40000000 /**< Endpoint Stall bit */
#define XUSB_EP_CFG_OUT_IN_MASK		0x20000000 /**< Endpoint OUT/IN cfg */
#define XUSB_EP_CFG_ISO_MASK		0x10000000 /**< Endpoint ISO config */
#define XUSB_EP_CFG_DATA_TOGGLE_MASK	0x08000000 /**< Endpoint Data toggle */
#define XUSB_EP_CFG_BUF_SELECT_MASK	0x04000000 /**< Endpoint Buff Select */
#define XUSB_EP_CFG_PACKET_SIZE_MASK	0x03ff8000 /**< Endpoint Packet Size */
#define XUSB_EP_CFG_BASE_MASK		0x00001fff /**< Endpoint Base Mask */

#define XUSB_EP_CFG_VALID_SHIFT		31	/**< Endpoint Valid bit Shift */
#define XUSB_EP_CFG_STALL_SHIFT		30	/**< Endpoint Stall bit Shift */
#define XUSB_EP_CFG_OUT_IN_SHIFT	29	/**< Endpoint OUT/IN config */
#define XUSB_EP_CFG_ISO_SHIFT		28	/**< Endpoint ISO config Shft */
#define XUSB_EP_CFG_DATA_TOGGLE_SHIFT	27	/**< Endpoint Data toggle  */
#define XUSB_EP_CFG_BUF_SELECT_SHIFT	26	/**< Endpoint Buf Select Shft */
#define XUSB_EP_CFG_PACKET_SIZE_SHIFT	15	/**< Endpoint Pkt Size Shift */

/* @} */

/*
 * Define the appropriate I/O access method to memory mapped I/O or
 * other interface if necessary
 */

#define XUsb_In32  XIo_In32
#define XUsb_Out32 XIo_Out32

/**************************** Type Definitions *******************************/

/***************** Macros (Inline Functions) Definitions *********************/

/****************************************************************************/
/**
* Read from the specified USB device register.
*
* @param	BaseAddress contains the base address of the device.
* @param	RegOffset contains the offset from the 1st register of the
*		device to select the specific register.
*
* @return	The value read from the register.
*
* @note		C-Style signature:
*		u32 XUsb_mReadReg(u32 BaseAddress, u32 RegOffset);
*
******************************************************************************/
#define XUsb_mReadReg(BaseAddress, RegOffset) \
	XUsb_In32((BaseAddress) + (RegOffset))

/***************************************************************************/
/**
* Write to the specified USB device register.
*
* @param	BaseAddress contains the base address of the device.
* @param	RegOffset contains the offset from the 1st register of the
*		device to select the specific register.
* @param	RegisterValue is the value to be written to the register.
*
* @return	None.
*
* @note		C-Style signature:
*		void XUsb_mWriteReg(u32 BaseAddress, u32 RegOffset,
*					u32 RegisterValue);
******************************************************************************/
#define XUsb_mWriteReg(BaseAddress, RegOffset, RegisterValue) \
	XUsb_Out32((BaseAddress) + (RegOffset), (RegisterValue))

#ifdef __cplusplus
}
#endif

#endif /* XUSB_L_H */
