//******************************************************************************
// *
// *       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
// *       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
// *       SOLUTIONS FOR XILINX DEVICES. BY PROVIDING THIS DESIGN, CODE,
// *       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
// *       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
// *       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
// *       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
// *       FOR YOUR IMPLEMENTATION. XILINX EXPRESSLY DISCLAIMS ANY
// *       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
// *       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
// *       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
// *       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// *       FOR A PARTICULAR PURPOSE.
// *
// *		Copyright(C) 2005 by xilinx, Inc. All rights reserved.              
// *
// *       (c) Copyright 2007 Xilinx Inc.
// *       All rights reserved.
//*
// ******************************************************************************/

#include "pch.h"
#include "..\intrface.h"
#ifdef XILINXDD_WMI_TRACE
#include "xilinxdd.tmh"
#endif

// global data
XILINXDD_DATA g_Data;

///////////////////////////////////////////////////////////////////////////////////////////////////
//  DriverEntry 
//      Installable driver initialization entry point.
//      This entry point is called directly by the I/O system.
//
//  Arguments:
//      IN  DriverObject
//              pointer to the driver object
//
//      IN  RegistryPath
//              pointer to a unicode string representing the path,
//              to driver-specific key in the registry.
//
//  Return Value:
//      Status
//
NTSTATUS DriverEntry(
    IN  PDRIVER_OBJECT  DriverObject,
    IN  PUNICODE_STRING RegistryPath
    )
{
    NTSTATUS    status;

    xilinxddDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"++");
    xilinxddDebugPrint(DBG_INIT, DBG_INFO, "Compiled at %s on %s", __TIME__, __DATE__);

#ifdef DBG
//    DbgBreakPoint();
#endif

#ifdef XILINXDD_WMI_TRACE 
    WPP_INIT_TRACING(DriverObject, RegistryPath);
#endif

    status = STATUS_SUCCESS;

    RtlZeroMemory(&g_Data, sizeof(XILINXDD_DATA));

    // save registry path for wmi
    g_Data.RegistryPath.Length = RegistryPath->Length;
    g_Data.RegistryPath.MaximumLength = RegistryPath->Length + sizeof(UNICODE_NULL);
    g_Data.RegistryPath.Buffer = (PWCHAR)ExAllocatePoolWithTag(
                                            PagedPool,
                                            g_Data.RegistryPath.MaximumLength,
                                            XILINXDD_POOL_TAG
                                            );

    if (g_Data.RegistryPath.Buffer == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

        xilinxddDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__": Failed to allocate memory for RegistryPath");

        return status;
    }

    RtlCopyUnicodeString(&g_Data.RegistryPath, RegistryPath);

    // detect current operating system
    if (IoIsWdmVersionAvailable(1, 0x30))
    {
        g_Data.WdmVersion = 0x0130;
    }
    else if (IoIsWdmVersionAvailable(1, 0x20))
    {
        g_Data.WdmVersion = 0x0120;
    }
    else if (IoIsWdmVersionAvailable(1, 0x10))
    {
        g_Data.WdmVersion = 0x0110;
    }
    else if (IoIsWdmVersionAvailable(1, 0x05))
    {
        g_Data.WdmVersion = 0x0105;
    }
    else
    {
        g_Data.WdmVersion = 0x0100;
    }

    // setup our dispatch function table in the driver object
    DriverObject->MajorFunction[IRP_MJ_PNP] = xilinxddPnpDispatch;
    DriverObject->MajorFunction[IRP_MJ_POWER] = xilinxddPowerDispatch;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = xilinxddCreateDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = xilinxddCloseDispatch;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = xilinxddDeviceIoControlDispatch;
    DriverObject->MajorFunction[IRP_MJ_READ] = xilinxddReadDispatch;
    DriverObject->MajorFunction[IRP_MJ_WRITE] = xilinxddWriteDispatch;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP] = xilinxddCleanupDispatch;
    DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = xilinxddSystemControlDispatch;

    DriverObject->DriverExtension->AddDevice = xilinxddAddDevice;
    DriverObject->DriverUnload = xilinxddUnload;

    xilinxddDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddAddDevice 
//      The PnP manager new device callback
//
//  Arguments:
//      IN  DeviceObject
//              pointer to a device object.
//
//      IN  PhysicalDeviceObject
//              pointer to a device object created by the
//              underlying bus driver.
//
//  Return Value:
//      Status
//
NTSTATUS xilinxddAddDevice(
    IN PDRIVER_OBJECT   DriverObject,
    IN PDEVICE_OBJECT   PhysicalDeviceObject
    )
{
    NTSTATUS                            status;
    PDEVICE_OBJECT                      deviceObject;
    PXILINXDD_DEVICE_EXTENSION   deviceExtension;
    POWER_STATE                         powerState;

    xilinxddDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"++: PDO %p", PhysicalDeviceObject);

    // Create our function device object.
    status = IoCreateDevice(
                DriverObject,
                sizeof (XILINXDD_DEVICE_EXTENSION),
                NULL,                    // no name
                FILE_DEVICE_UNKNOWN,
                FILE_DEVICE_SECURE_OPEN, // Do not use if binary intended for 9x
                FALSE,
                &deviceObject
                );

    if (!NT_SUCCESS(status)) 
    {
        xilinxddDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoCreateDevice STATUS %x", status);

        return status;
    }

    xilinxddDebugPrint(DBG_INIT, DBG_INFO, __FUNCTION__": Created FDO %p", deviceObject);

    // Initialize the device extension.
    deviceExtension = (PXILINXDD_DEVICE_EXTENSION)deviceObject->DeviceExtension;

    // Zero the memory
    RtlZeroMemory(deviceExtension, sizeof(XILINXDD_DEVICE_EXTENSION));

    // save the PDO pointer
    deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;

    // save our device object
    deviceExtension->DeviceObject = deviceObject;

    // Initialize the powering down flag
    deviceExtension->bPowerStop = FALSE;

    // set RemoveCount to 1. Transition to zero
    // means IRP_MN_REMOVE_DEVICE was received
    deviceExtension->RemoveCount = 1;

    // Initialize Remove event
    KeInitializeEvent(&deviceExtension->RemoveEvent, NotificationEvent, FALSE);

    // initialize io lock
    xilinxddInitializeIoLock(&deviceExtension->IoLock, deviceObject);

    deviceExtension->PnpState = PnpStateNotStarted;
    deviceExtension->PreviousPnpState = PnpStateNotStarted;

    // allocate WorkItem for power processing
    deviceExtension->PowerWorkItem = IoAllocateWorkItem(deviceExtension->DeviceObject);
    if (deviceExtension->PowerWorkItem == NULL)
    {
        xilinxddDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. Failed to allocated PowerWorkItem");

        IoDeleteDevice(deviceObject);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // Initialize the device object flags

    // All WDM drivers are supposed to set this flag.  Devices
    // requiring a large amount of current at statup set the 
    // DO_POWER_INRUSH flag instead.  These flags are mutually
    // exclusive.
    if (PhysicalDeviceObject->Flags & DO_POWER_PAGABLE) 
    {
        deviceObject->Flags |= DO_POWER_PAGABLE;
    }

    // This flag sets the buffering method for reads and writes
    // to METHOD_DIRECT.  IOCTLs are handled by IO control codes
    // independent of the value of this flag.
    deviceObject->Flags |= DO_DIRECT_IO;

    // Typically, (almost always) the function driver for a device is the
    // power policy owner for that device.  In some cases
    // another driver or system component may assume this role.

    // Set the initial device power state of the device to powered down.
    deviceExtension->DevicePowerState = PowerDeviceD3;
    powerState.DeviceState = deviceExtension->DevicePowerState;

    // the initial system power state must be working or we wouldn't be here.
    deviceExtension->SystemPowerState = PowerSystemWorking;

    // Call PoSetPowerState to notify the power manager of our initial power state.
    PoSetPowerState(deviceObject, DevicePowerState, powerState);

    // Attach our device to the device stack and get the device object to 
    // which we send down IRPs to.
    deviceExtension->LowerDeviceObject = IoAttachDeviceToDeviceStack(
                                            deviceObject,
                                            PhysicalDeviceObject
                                            );

    if (deviceExtension->LowerDeviceObject == NULL) 
    {
        xilinxddDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoAttachDeviceToDeviceStack failed");

        IoFreeWorkItem(deviceExtension->PowerWorkItem);
        IoDeleteDevice(deviceObject);
        return STATUS_DEVICE_REMOVED;
    }

    deviceExtension->Pipe7Urb = (PURB)ExAllocatePoolWithTag(
                                NonPagedPool, 
                                sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
                                XILINXDD_POOL_TAG
                                );

    if (deviceExtension->Pipe7Urb == NULL)
    {
    }

    deviceExtension->Pipe7Irp = IoAllocateIrp(
                                    deviceExtension->LowerDeviceObject->StackSize, 
                                    FALSE
                                    );

    // Initialize our device's interface
    status = IoRegisterDeviceInterface (
                PhysicalDeviceObject,
                (LPGUID)&GUID_DEVINTERFACE_XILINXDD,
                NULL,
                &deviceExtension->InterfaceName
                );

    if (!NT_SUCCESS(status)) 
    {
        xilinxddDebugPrint(DBG_INIT, DBG_ERR, __FUNCTION__"--. IoRegisterDeviceInterface STATUS %x", status);

        IoDetachDevice(deviceExtension->LowerDeviceObject);
        IoFreeWorkItem(deviceExtension->PowerWorkItem);
        IoDeleteDevice(deviceObject);
        return status;
    }

    // increment device instance count
    ++g_Data.InstanceCount;

    // We are all done, so we need to clear the
    // DO_DEVICE_INITIALIZING flag.  This must be our
    // last action in AddDevice
    deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    xilinxddDebugPrint(DBG_INIT, DBG_TRACE, __FUNCTION__"--. xilinxddAddDevice--. STATUS %x", status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddStartDevice
//      Start device handler, sets up resources
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  Irp
//              pointer to the Start IRP
//
//  Return Value:
//      NT status code
//
NTSTATUS xilinxddStartDevice(
    IN  PXILINXDD_DEVICE_EXTENSION    DeviceExtension,
    IN  PIRP                            Irp
    )
{
    NTSTATUS                        status;
    PURB                            urb;
    PUSB_DEVICE_DESCRIPTOR          deviceDescriptor;
    PUSB_CONFIGURATION_DESCRIPTOR   configDescriptor;
    USHORT                          size;
    PUSBD_INTERFACE_LIST_ENTRY      interfaceBuffer;
    PUSBD_INTERFACE_LIST_ENTRY      interfaceList;
    UCHAR                           numInterfaces;
    UCHAR                           index;
    PUSB_INTERFACE_DESCRIPTOR       interfaceDescriptor;
    PUSBD_INTERFACE_INFORMATION     interfaceInfo;
    PUSB_ENDPOINT_DESCRIPTOR        epDescriptor;

    xilinxddDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    // we will use do{}while(FALSE); structure
    // to help with resource cleanup
    urb = NULL;
    deviceDescriptor = NULL;
    configDescriptor = NULL;
    interfaceBuffer = NULL;

    do
    {
        // we need to allocate an URB to talk
        // to usb bus driver
        urb = (PURB)ExAllocatePoolWithTag(
                        NonPagedPool,
                        sizeof(URB),
                        XILINXDD_POOL_TAG
                        );

        if (urb == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        // first let's read device descriptor
        deviceDescriptor = (PUSB_DEVICE_DESCRIPTOR)ExAllocatePoolWithTag(
                                                    NonPagedPool,
                                                    sizeof(USB_DEVICE_DESCRIPTOR),
                                                    XILINXDD_POOL_TAG
                                                    );
        if (deviceDescriptor == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        UsbBuildGetDescriptorRequest(
            urb,
            (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
            USB_DEVICE_DESCRIPTOR_TYPE,
            0,
            0,
            deviceDescriptor,
            NULL,
            sizeof(USB_DEVICE_DESCRIPTOR),
            NULL
            );

        status = xilinxddSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
            break;
        }

        // We have to read configuration descriptor twice.
        // First time to obtain its full length and second time
        // to read full configuration descriptor
        configDescriptor = (PUSB_CONFIGURATION_DESCRIPTOR)ExAllocatePoolWithTag(
                                                            NonPagedPool,
                                                            sizeof(USB_CONFIGURATION_DESCRIPTOR),
                                                            XILINXDD_POOL_TAG
                                                            );

        if (configDescriptor == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        UsbBuildGetDescriptorRequest(
            urb,
            (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
            USB_CONFIGURATION_DESCRIPTOR_TYPE,
            0,
            0,
            configDescriptor,
            NULL,
            sizeof(USB_CONFIGURATION_DESCRIPTOR),
            NULL
            );

        status = xilinxddSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
            break;
        }

        size = configDescriptor->wTotalLength;
        ExFreePool(configDescriptor);

        configDescriptor = (PUSB_CONFIGURATION_DESCRIPTOR)ExAllocatePoolWithTag(
                                                            NonPagedPool,
                                                            size,
                                                            XILINXDD_POOL_TAG
                                                            );

        if (configDescriptor == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        UsbBuildGetDescriptorRequest(
            urb,
            (USHORT)sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST),
            USB_CONFIGURATION_DESCRIPTOR_TYPE,
            0,
            0,
            configDescriptor,
            NULL,
            size,
            NULL
            );

        status = xilinxddSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
            break;
        }

        // now that we have configuration descriptor lets find our interface
        // and enable it
        numInterfaces = configDescriptor->bNumInterfaces;
        interfaceBuffer = interfaceList =
            (PUSBD_INTERFACE_LIST_ENTRY)ExAllocatePoolWithTag(
                                            NonPagedPool,
                                            sizeof(USBD_INTERFACE_LIST_ENTRY) * (numInterfaces + 1),
                                            XILINXDD_POOL_TAG
                                            );

        if (interfaceBuffer == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        for (index = 0; index < numInterfaces; ++index)
        {
            interfaceDescriptor = USBD_ParseConfigurationDescriptorEx(
                                    configDescriptor,
                                    configDescriptor,
                                    index,
                                    0,
                                    -1,
                                    -1,
                                    -1
                                    );

            if (interfaceDescriptor != NULL)
            {
                interfaceList->InterfaceDescriptor = interfaceDescriptor;
                interfaceList->Interface = NULL;

                ++interfaceList;
            }
        }

        interfaceList->InterfaceDescriptor = NULL;
        interfaceList->Interface = NULL;

        ExFreePool(urb);
        urb = USBD_CreateConfigurationRequestEx(configDescriptor, interfaceBuffer);
        if (urb == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        interfaceInfo = &urb->UrbSelectConfiguration.Interface;

        // verify configuration
        interfaceDescriptor = interfaceBuffer->InterfaceDescriptor;
        if (interfaceDescriptor->bNumEndpoints != 7)
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)interfaceDescriptor;

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x1) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 512))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[0].MaximumTransferSize = 65536;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x2) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 512))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[1].MaximumTransferSize = 65536;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x3) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 512))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[2].MaximumTransferSize = 65536;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x84) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 512))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[3].MaximumTransferSize = 65536;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x85) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 512))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[4].MaximumTransferSize = 65536;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x86) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_BULK) ||
            (epDescriptor->wMaxPacketSize != 512))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[5].MaximumTransferSize = 65536;
            ++epDescriptor;
        }

        epDescriptor = (PUSB_ENDPOINT_DESCRIPTOR)USBD_ParseDescriptors(
                                                    configDescriptor,
                                                    configDescriptor->wTotalLength,
                                                    epDescriptor,
                                                    USB_ENDPOINT_DESCRIPTOR_TYPE
                                                    );

        if ((epDescriptor == NULL) ||
            (epDescriptor->bEndpointAddress != 0x87) ||
            (epDescriptor->bmAttributes != USB_ENDPOINT_TYPE_INTERRUPT) ||
            (epDescriptor->wMaxPacketSize != 16))
        {
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
            break;
        }
        else
        {
            interfaceInfo->Pipes[6].MaximumTransferSize = 4096;
            ++epDescriptor;
        }

        status = xilinxddSubmitUrbSynch(DeviceExtension, urb);
        if (!NT_SUCCESS(status))
        {
            break;
        }

        DeviceExtension->InterfaceInformation =
            (PUSBD_INTERFACE_INFORMATION)ExAllocatePoolWithTag(
                                            NonPagedPool,
                                            interfaceInfo->Length,
                                            XILINXDD_POOL_TAG
                                            );

        if (DeviceExtension->InterfaceInformation == NULL)
        {
            status = STATUS_INSUFFICIENT_RESOURCES;
            break;
        }

        RtlCopyMemory(DeviceExtension->InterfaceInformation, interfaceInfo, interfaceInfo->Length);

        DeviceExtension->ConfigHandle = urb->UrbSelectConfiguration.ConfigurationHandle;
        DeviceExtension->ConfigDescriptor = configDescriptor;
        configDescriptor = NULL;
    }
    while (FALSE);

    if (urb != NULL)
    {
        ExFreePool(urb);
    }

    if (deviceDescriptor != NULL)
    {
        ExFreePool(deviceDescriptor);
    }

    if (configDescriptor != NULL)
    {
        ExFreePool(configDescriptor);
    }

    if (interfaceBuffer != NULL)
    {
        ExFreePool(interfaceBuffer);
    }

    xilinxddDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddFreeResources
//      This routine returns all the USB resources acquired during
//      device startup.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      NT status code
//
NTSTATUS xilinxddFreeResources(
    IN  PXILINXDD_DEVICE_EXTENSION DeviceExtension
    )
{
    NTSTATUS    status = STATUS_SUCCESS;
    PURB        urb;

    xilinxddDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"++");

    if (DeviceExtension->Pipe7Urb != NULL)
    {
        ExFreePool(DeviceExtension->Pipe7Urb);
        DeviceExtension->Pipe7Urb = NULL;
    }

    if (DeviceExtension->Pipe7Irp != NULL)
    {
        IoFreeIrp(DeviceExtension->Pipe7Irp);
        DeviceExtension->Pipe7Irp = NULL;
    }
    if (DeviceExtension->ConfigDescriptor != NULL)
    {
        if (DeviceExtension->PnpState != PnpStateSurpriseRemoved)
        {
            urb = (PURB)ExAllocatePoolWithTag(
                            NonPagedPool,
                            sizeof(struct _URB_SELECT_CONFIGURATION),
                            XILINXDD_POOL_TAG
                            );

            if (urb != NULL)
            {
                UsbBuildSelectConfigurationRequest(
                    urb,
                    (USHORT)sizeof(struct _URB_SELECT_CONFIGURATION),
                    NULL
                    );

                status = xilinxddSubmitUrbSynch(DeviceExtension, urb);

                ExFreePool(urb);
            }
            else
            {
                status = STATUS_INSUFFICIENT_RESOURCES;
            }
        }

        if (DeviceExtension->ConfigDescriptor != NULL)
        {
            ExFreePool(DeviceExtension->ConfigDescriptor);
            DeviceExtension->ConfigDescriptor = NULL;
        }
    }

    if (DeviceExtension->InterfaceInformation != NULL)
    {
        ExFreePool(DeviceExtension->InterfaceInformation);
        DeviceExtension->InterfaceInformation = NULL;
    }

    xilinxddDebugPrint(DBG_PNP, DBG_TRACE, __FUNCTION__"--");

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddCreateDispatch
//      Dispatch routine for IRP_MJ_CREATE requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the create IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS xilinxddCreateDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PXILINXDD_DEVICE_EXTENSION     deviceExtension;
    PFILE_OBJECT                    fileObject;
    PIO_STACK_LOCATION              irpStack;
    ULONG                           pipeNum;
    PUSBD_INTERFACE_INFORMATION     interface;
    NTSTATUS                        status;

    xilinxddDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__": IRP %p", Irp);

    //
    // initialize variables
    //
    irpStack = IoGetCurrentIrpStackLocation(Irp);
    fileObject = irpStack->FileObject;
    deviceExtension = (PXILINXDD_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    if (!xilinxddAcquireRemoveLock(deviceExtension))
    {
        status = STATUS_NO_SUCH_DEVICE;

        Irp->IoStatus.Status = status;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);

        xilinxddDebugPrint(DBG_CREATECLOSE, DBG_WARN, __FUNCTION__"$$--. IRP %p, STATUS %x", Irp, status);

        goto FuncExit;
    }



    if(deviceExtension->InterfaceInformation)
    {

        interface = deviceExtension->InterfaceInformation;
    } else
    {

        xilinxddDebugPrint(DBG_CREATECLOSE, DBG_WARN, __FUNCTION__"Interface not found");

        status = STATUS_INVALID_DEVICE_STATE;
        goto FuncExit;
    }

    //
    // FsContext is Null for the device
    //
    if(fileObject)
    {

        fileObject->FsContext = NULL; 
    } else
    {

        status = STATUS_INVALID_PARAMETER;
        goto FuncExit;
    }

    pipeNum = xilinxdd_PipeWIthName(DeviceObject, &fileObject->FileName);

    if(pipeNum == -1)
    {
        status = STATUS_INVALID_PARAMETER;
        goto FuncExit;
    }

    if(pipeNum < interface->NumberOfPipes)
    {
        fileObject->FsContext = &interface->Pipes[pipeNum -1];
    }

    // increment open count
    InterlockedIncrement(&deviceExtension->OpenHandleCount);

    status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    xilinxddReleaseRemoveLock(deviceExtension);

    xilinxddDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

FuncExit:
    return status;
}

/*++
 
Routine Description:

    This routine will convert the last WCHAR of the file name to a pipe number
    from 1 to 9

Arguments:

    DeviceObject - pointer to DeviceObject
    FileName - string pipe name

Return Value:

    The Pipe number 

--*/

LONG
xilinxdd_PipeWIthName(
                    IN PDEVICE_OBJECT  DeviceObject,
                    IN PUNICODE_STRING FileName
                    )
{
    LONG                            ix;
    ULONG                           uval; 
    ULONG                           nameLength;
    ULONG                           umultiplier;
    PXILINXDD_DEVICE_EXTENSION     deviceExtension;
    LONG                            pipeNum;
    //
    // initialize variables
    //
    pipeNum = -1;
    //
    // typedef WCHAR *PWSTR;
    //
    nameLength = (FileName->Length / sizeof(WCHAR));
    deviceExtension = (PXILINXDD_DEVICE_EXTENSION) DeviceObject->DeviceExtension;


    if(nameLength != 0)
    {

        //
        // Parse the pipe#
        //
        ix = nameLength - 1;

        if((FileName->Buffer[ix] >= (WCHAR) '1') && (FileName->Buffer[ix] <= (WCHAR) '7'))
        {
            // found our pipe
            pipeNum = FileName->Buffer[ix] - (WCHAR) '0';
        }
    }

    return(pipeNum);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddCloseDispatch
//      Dispatch routine for IRP_MJ_CLOSE requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the close IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS xilinxddCloseDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PXILINXDD_DEVICE_EXTENSION     deviceExtension;
    NTSTATUS                        status;

    xilinxddDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PXILINXDD_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    InterlockedDecrement(&deviceExtension->OpenHandleCount);

    xilinxddDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p, STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddCleanupDispatch
//      Dispatch routine for IRP_MJ_CLEANUP requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to the device object for our device
//
//      IN  Irp
//              the cleanup IRP
//
//  Return Value:
//      NT status code.
//
NTSTATUS xilinxddCleanupDispatch(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    PXILINXDD_DEVICE_EXTENSION    deviceExtension;
    NTSTATUS                status;
    PIO_STACK_LOCATION      irpStack;

    xilinxddDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"++. IRP %p", Irp);

    deviceExtension = (PXILINXDD_DEVICE_EXTENSION)DeviceObject->DeviceExtension;

    irpStack = IoGetCurrentIrpStackLocation(Irp);
    xilinxddFlushQueues(deviceExtension, irpStack->FileObject);

    status = STATUS_SUCCESS;

    Irp->IoStatus.Information = 0;
    Irp->IoStatus.Status = status;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    xilinxddDebugPrint(DBG_CREATECLOSE, DBG_TRACE, __FUNCTION__"--. IRP %p STATUS %x", Irp, status);

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddUnload
//      Driver Unload routine.
//
//  Arguments:
//      IN  DriverObject
//              pointer to the driver object
//
//  Return Value:
//      none
//
VOID xilinxddUnload(
    IN  PDRIVER_OBJECT  DriverObject
    )
{
    xilinxddDebugPrint(DBG_UNLOAD, DBG_TRACE, __FUNCTION__"++");

    // The device object(s) should be NULL now
    // (since we unload, all the devices objects associated with this
    // driver must be deleted.
    ASSERT(DriverObject->DeviceObject == NULL);

    // We should not be unloaded until all the devices we control
    // have been removed from our queue.

    // release memory block allocated for registry path
    if (g_Data.RegistryPath.Buffer != NULL)
    {
        ExFreePool(g_Data.RegistryPath.Buffer);
        g_Data.RegistryPath.Buffer = NULL;
    }

    xilinxddDebugPrint(DBG_UNLOAD, DBG_TRACE, __FUNCTION__"--");

#ifdef XILINXDD_WMI_TRACE
    WPP_CLEANUP(DriverObject);
#endif

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddIsStoppable
//      This routine determines whether the device can be safely stopped. 
//      In our particular case, we'll assume we can always stop the device.
//      A device might fail the request if it doesn't have a queue for the
//      requests that might come or if it was notified that it is in the 
//      paging path.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      returns TRUE if the device is stoppable, FALSE otherwise
//
BOOLEAN xilinxddIsStoppable(
    IN  PXILINXDD_DEVICE_EXTENSION    DeviceExtension
    )
{
    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddIsRemovable
//      This routine determines whether the device can be safely removed. 
//      A device shouldn't be removed if, for example, it has open handles or
//      removing the device could result in losing data (plus the reasons
//      mentioned in the xilinxddIsStoppable function comments). The PnP manager on 
//      Windows 2000 fails on its own any attempt to remove, if there any 
//      open handles to the device.  However on Win9x, the driver must keep 
//      count of open handles and fail QueryRemove if there are any open 
//      handles.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      returns TRUE if the device is stoppable, FALSE otherwise
//
BOOLEAN xilinxddIsRemovable(
    IN  PXILINXDD_DEVICE_EXTENSION    DeviceExtension
    )
{
    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddSubmitIrpSyncComplete
//      Completion routine for sync IRP requests.
//
//  Arguments:
//      IN  DeviceObject
//              pointer to our device object
//
//      IN  Irp
//              pointer to the PnP IRP
//
//      IN  Context
//              our event used to signal IRP completion
//
//  Return Value:
//      STATUS_MORE_PROCESSING_REQUIRED
//
NTSTATUS  xilinxddSubmitIrpSyncComplete(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp,
    IN  PVOID           Context
    )
{
    PKEVENT event = (PKEVENT)Context;

    // If the lower driver didn't return STATUS_PENDING, we don't need to 
    // set the event because we won't be waiting on it. 
    if (Irp->PendingReturned) 
    {
        KeSetEvent(event, IO_NO_INCREMENT, FALSE);
    }

    return STATUS_MORE_PROCESSING_REQUIRED;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddSubmitIrpSync
//      Sends the given IRP down the stack to the next lower driver and 
//      waits in a synchronous fashion for the IRP to complete
//
//  Arguments:
//      IN  DeviceObject
//              Pointer to device object for our device
//
//      IN  Irp
//              IRP to send down
//
//  Return Value:
//      NT status code
//
NTSTATUS xilinxddSubmitIrpSync(
    IN  PDEVICE_OBJECT  DeviceObject,
    IN  PIRP            Irp
    )
{
    KEVENT   event;
    NTSTATUS status;

    KeInitializeEvent(&event, NotificationEvent, FALSE);

    IoCopyCurrentIrpStackLocationToNext(Irp);

    IoSetCompletionRoutine(
        Irp,
        xilinxddSubmitIrpSyncComplete,
        &event,
        TRUE,
        TRUE,
        TRUE
        );

    status = IoCallDriver(DeviceObject, Irp);

    // Wait for lower drivers to be done with the Irp.
    // Important thing to note here is when you allocate
    // memory for an event in the stack you must do a
    // KernelMode wait instead of UserMode to prevent
    // the stack from getting paged out.

    if (status == STATUS_PENDING) 
    {
       KeWaitForSingleObject(
           &event,
           Executive,
           KernelMode,
           FALSE,
           NULL
           );

       status = Irp->IoStatus.Status;
    }

    return status;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddAcquireRemoveLock
//      Acquires remove lock.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      FALSE if remove device pending, TRUE otherwise.
//
BOOLEAN xilinxddAcquireRemoveLock(
    IN  PXILINXDD_DEVICE_EXTENSION    DeviceExtension
    )
{
    LONG    count;

    count = InterlockedIncrement(&DeviceExtension->RemoveCount);

    ASSERT(count > 0);

    if (DeviceExtension->PnpState == PnpStateRemoved)
    {
        xilinxddReleaseRemoveLock(DeviceExtension);

        return FALSE;
    }

    return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddReleaseRemoveLock
//      Releases remove lock.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      None.
//
VOID xilinxddReleaseRemoveLock(
    IN  PXILINXDD_DEVICE_EXTENSION    DeviceExtension
    )
{
    LONG    count;

    count = InterlockedDecrement(&DeviceExtension->RemoveCount);

    ASSERT(count >= 0);

    if (count == 0)
    {
        KeSetEvent(&DeviceExtension->RemoveEvent, IO_NO_INCREMENT, FALSE);
    }

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddWaitForSafeRemove
//      Waits for all remove locks to be released
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      None.
//
//  Comment:
//      This routine should be called with no remove locks held 
//      by the calling thread
//
VOID xilinxddWaitForSafeRemove(
    IN  PXILINXDD_DEVICE_EXTENSION    DeviceExtension
    )
{
    DeviceExtension->PnpState = PnpStateRemoved;
    xilinxddReleaseRemoveLock(DeviceExtension);

    KeWaitForSingleObject(
        &DeviceExtension->RemoveEvent, 
        Executive, 
        KernelMode, 
        FALSE, 
        NULL
        );

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddStallQueues
//      Pauses all of the queues, and waits for them to get to paused state.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      none
//
VOID xilinxddStallQueues(
    IN  PXILINXDD_DEVICE_EXTENSION DeviceExtension
    )
{
    ULONG   index;

    xilinxddDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    // stall IRP processing
    xilinxddLockIo(&DeviceExtension->IoLock);
    xilinxddWaitForStopIo(&DeviceExtension->IoLock);
    xilinxddDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddRestartQueues
//      Restarts all paused queues.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//  Return Value:
//      none
//
VOID xilinxddRestartQueues(
    IN  PXILINXDD_DEVICE_EXTENSION   DeviceExtension
    )
{
    ULONG   index;

    xilinxddDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");
    xilinxddUnlockIo(&DeviceExtension->IoLock);
    xilinxddDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}
///////////////////////////////////////////////////////////////////////////////////////////////////
//  xilinxddFlushQueues
//      Flush oustanding IRPs for closed file object.
//
//  Arguments:
//      IN  DeviceExtension
//              our device extension
//
//      IN  FileObject
//              about to be closed file object
//
//  Return Value:
//      none
//
VOID xilinxddFlushQueues(
    IN  PXILINXDD_DEVICE_EXTENSION   DeviceExtension,
    IN  PFILE_OBJECT            FileObject
    )
{
    ULONG   index;

    xilinxddDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"++");

    xilinxddFlushPendingIo(&DeviceExtension->IoLock, FileObject);

    xilinxddDebugPrint(DBG_IO, DBG_TRACE, __FUNCTION__"--");

    return;
}
