//******************************************************************************
// *
// *       XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
// *       AS A COURTESY TO YOU, SOLELY FOR USE IN DEVELOPING PROGRAMS AND
// *       SOLUTIONS FOR XILINX DEVICES. BY PROVIDING THIS DESIGN, CODE,
// *       OR INFORMATION AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE,
// *       APPLICATION OR STANDARD, XILINX IS MAKING NO REPRESENTATION
// *       THAT THIS IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
// *       AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
// *       FOR YOUR IMPLEMENTATION. XILINX EXPRESSLY DISCLAIMS ANY
// *       WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
// *       IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
// *       REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
// *       INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// *       FOR A PARTICULAR PURPOSE.
// *
// *		Copyright(C) 2005 by xilinx, Inc. All rights reserved.              
// *
// *       (c) Copyright 2007 Xilinx Inc.
// *       All rights reserved.
//*
// ******************************************************************************/

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <commctrl.h>
#include <tchar.h>
#include <setupapi.h>
#include <initguid.h>
#include <winioctl.h>
#include <process.h>
#include <assert.h>
#include <dbt.h>
#include "usb.h"

#include "..\intrface.h"

#define MAX_STRING_LENGTH       500

WCHAR *DevicePath;
HANDLE epHandles[8];

// Find our USB device and get the path to it.
// Save that for later open of each endpoint, by appending the
// endpoint number to the path we find. The path name these days is
// UGLY!!!!
DWORD USBInit(void)
{
    DWORD                               results;
    BOOL                                boolResult;
    int                                 i;
    HDEVINFO                            hDeviceInfo;
    DWORD                               bufferSize;
    SP_DEVICE_INTERFACE_DATA            interfaceData;
    PSP_DEVICE_INTERFACE_DETAIL_DATA    deviceDetail;

    for (i = 0; i < 8; i++)
    {
        epHandles[i] = INVALID_HANDLE_VALUE;
    }

    // Find all devices that have our interface
    hDeviceInfo = SetupDiGetClassDevs(
                                        (LPGUID)&GUID_DEVINTERFACE_XILINXDD,
                                        NULL,
                                        NULL,
                                        DIGCF_PRESENT | DIGCF_DEVICEINTERFACE
                                     );
    if (hDeviceInfo == INVALID_HANDLE_VALUE)
    {
        printf("Could not find class\n");
        results = GetLastError();
        goto FuncExit;
    }

    // Setup the interface data struct
    interfaceData.cbSize = sizeof(interfaceData);

    boolResult = SetupDiEnumDeviceInterfaces(
                                            hDeviceInfo,
                                            NULL,

                                            (LPGUID)
                                            &GUID_DEVINTERFACE_XILINXDD,
                                            0,
                                            &interfaceData
                                            );
    if(boolResult == FALSE)
    {
        printf("Could not find Device I/F\n");
        results = GetLastError();
        goto FuncExit;
    }
    // Found our device instance
    boolResult = SetupDiGetDeviceInterfaceDetail(
                                                hDeviceInfo,
                                                &interfaceData,
                                                NULL,
                                                0,
                                                &bufferSize,
                                                NULL
                                               );
    if(boolResult == FALSE)
    {
        results = GetLastError();
        if (results != ERROR_INSUFFICIENT_BUFFER)
        {
            printf("Could not setup I/F\n");
            goto FuncExit;
        }
    }

    // Allocate a big enough buffer to get detail data
    deviceDetail = (PSP_DEVICE_INTERFACE_DETAIL_DATA)malloc(bufferSize);
    if (deviceDetail == NULL)
    {
        results = GetLastError();
        goto FuncExit;
    }

    DevicePath = (WCHAR *)malloc(bufferSize * 4);
    if (DevicePath == NULL)
    {
        results = GetLastError();
        goto FuncExit;
    }

    // Setup the device interface struct
    deviceDetail->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    // Try again to get the device interface detail info
    boolResult = SetupDiGetDeviceInterfaceDetail(
                                                hDeviceInfo,
                                                &interfaceData,
                                                deviceDetail,
                                                bufferSize,
                                                NULL,
                                                NULL
                                               );
    if (boolResult == 0)
    {
        printf("Could not get device I/F detail\n");
        results = GetLastError();
        free(deviceDetail);
        goto FuncExit;
    }
    wcsncpy(DevicePath, deviceDetail->DevicePath, bufferSize);
    wprintf(L"USB: Found device = %s\n", DevicePath);
    free(deviceDetail);
    SetupDiDestroyDeviceInfoList(hDeviceInfo);
    results = ERROR_SUCCESS;

FuncExit:
    return(results);
}

/* Open an endpoint for I/O, the handle returned can be used
    in the other routines below for I/O
*/

DWORD USBOpen(int end_point)
{
    DWORD results  = ERROR_SUCCESS;
    WCHAR ep[16];
    WCHAR newPath[512];

    if(epHandles[end_point] != INVALID_HANDLE_VALUE)
    {
        // Handle is already open
        results = ERROR_SUCCESS;
        goto FuncExit;
    }

    wcsncpy(newPath, DevicePath, 512);
    _itow(end_point, ep, 10);

    // Open the path
    // deviceDetail->DevicePath
    // Free our allocated buffer

    // Open handle to device
    wcscat(newPath, L"\\");
    wcscat(newPath, ep);

    epHandles[end_point] = CreateFile(
                                        newPath,
                                        (GENERIC_READ | GENERIC_WRITE),
                                        (FILE_SHARE_READ | FILE_SHARE_WRITE),
                                        NULL,
                                        OPEN_EXISTING,
                                        0,
                                        0
                                      );

    if (epHandles[end_point] == INVALID_HANDLE_VALUE)
    {
        results = GetLastError();
        goto FuncExit;
    }

    wprintf(L"USB: Opened path = %s\n", newPath);

FuncExit:
    return(results);
}

DWORD  USBClose(int end_point)
{
    DWORD results;

    if(!CloseHandle(epHandles[end_point]))
    {
        results = GetLastError();
    } else
    {
        epHandles[end_point] = INVALID_HANDLE_VALUE;
        results = ERROR_SUCCESS;
    }

    return(results);
}

DWORD USBRead(int end_point, char *ptr, int len)
{
    DWORD results;

    if(epHandles[end_point] == INVALID_HANDLE_VALUE)
    {
        results = ERROR_INVALID_HANDLE;
        goto FuncExit;
    }

    if(!ReadFile(epHandles[end_point], (void *)ptr, len, &results, NULL))
    {
        results = GetLastError();
        goto FuncExit;
    }

FuncExit:
    return(results);
}

DWORD USBWrite(int end_point, char *ptr, int len)
{
    DWORD results;

    if(epHandles[end_point] == INVALID_HANDLE_VALUE)
    {
        results = ERROR_INVALID_HANDLE;
        goto FuncExit;
    }

    if(!WriteFile(epHandles[end_point], (void *)ptr, len, &results, NULL))
    {
        results = GetLastError();
        goto FuncExit;
    }

FuncExit:
    return(results);
}

DWORD USBIoctl(int end_point, DWORD request, void *inbuffer, DWORD inbuffersize,
                void *outbuffer, DWORD outbuffersize)
{
    DWORD results;

    if(epHandles[end_point] == INVALID_HANDLE_VALUE)
    {
        results = ERROR_INVALID_HANDLE;
        goto FuncExit;
    }

    if(!DeviceIoControl(epHandles[end_point], request, inbuffer, inbuffersize,
                        outbuffer, outbuffersize, &results, NULL))
    {
        results = GetLastError();
        goto FuncExit;
    }
FuncExit:
    return(results);
}



